import java.awt.*;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.ListIterator;
import java.util.Random;

/**
 * @author bachmaer
 *         <p/>
 *         Double buffered drawing canvas for displaying and storing game state.
 */
public class Sector extends Canvas implements Runnable {
    // Holds alien SpaceCraft in the sector
    ArrayList inSector = new ArrayList();

    // Holds obstacles in the sector
    ArrayList obstacles = new ArrayList();

    // Reference to the users own ship
    OwnSpaceCraft ownShip = null;

    // Back buffer for rendering
    Image offImage = null;
    Graphics offGraphics = null;
    Dimension offDimension = null;

    // delay between frames in milliseconds
    int delay;

    // Rendering thread
    Thread animatorThread;

    // Displayed size
    int sizeX = Constants.MAX_SECTOR_X;
    int sizeY = Constants.MAX_SECTOR_Y;

    // Random number generator for making star positions
    static Random rand = new Random();

    // Array of 2D points to use for displaying stars
    Point[] starPositions = new Point[250];

    Obstacle x = new Obstacle();

    private int ownPort;

    /**
     * Default constructor. Sets rendering speed for 20 fps.
     */
    public Sector(int ownPort) {

        this(ownPort, 20);

    } // end default construtor


    /**
     * Initializes the drawing canvas. Sets the size, frame rate, and background
     * color. Creates event handlers and start the animation thread.
     *
     * @param fps number of frames to be rendered per second
     */
    public Sector(int ownPort, int fps) {
        //travis added this
        this.ownPort = ownPort;

        this.setSize(sizeX, sizeY);

        delay = 1000 / fps;

        this.setBackground(Color.black);

        createStars();

        this.addKeyListener(new keyboardHandler());

        this.addComponentListener(new sizeChangeListener());

        animatorThread = new Thread(this);
        animatorThread.start();

    } // end Sector constructor


    /**
     * Checks for a collision between the users spacecraft and all aliens in the
     * sector. Returns the subject of the collision if one occured. If the
     * collision was headon. Ownship is set to
     * null.
     *
     * @return ship collided with if a collision occured. null otherwise.
     */
    public synchronized SpaceCraft collisionCheck() {
        int craftX = ownShip.getXPosition();
        int craftY = ownShip.getYPosition();

        double diffX, diffY;

        SpaceCraft target;

        ListIterator iter = inSector.listIterator();

        while (iter.hasNext()) {

            target = (SpaceCraft) (iter.next());

            diffX = craftX - target.getXPosition();
            diffY = craftY - target.getYPosition();

            if (Math.sqrt(diffX * diffX + diffY * diffY) <= 10.0) {

                System.out.println("Collision");

                inSector.remove(target);

                // Check for head on collision
                if (Math.abs(ownShip.getHeading() - target.getHeading()) == 2) {

                    ownShip = null;
                }

                return target;

            }
        } // end while

        return null;

    } // end collisionCheck


    /**
     * If the input parameter SpaceCraft is already in the sector its heading
     * and/or postion are updated. If the SpaceCrate is not found in the sector,
     * it is added.
     *
     * @param craft to be undated or added
     */
    public synchronized void updateAndAddSpaceCraft(SpaceCraft craft) {
        SpaceCraft target;
        ListIterator iter = inSector.listIterator();

        while (iter.hasNext()) {

            target = (SpaceCraft) iter.next();

            if (craft.equals(target)) {

                target.setHeading(craft.heading);
                target.setX(craft.xPosition);
                target.setY(craft.yPosition);

                return;
            }
        }

        inSector.add(craft);

        return;

    } // end updateAndAdd


    /**
     * Check to see if an obstacle is in front of ownShip.
     *
     * @param true if the position is clear, false otherwise
     */
    public synchronized boolean obstacleClear() {
        int xObs = ownShip.getXPosition();
        int yObs = ownShip.getYPosition();

        switch (ownShip.getHeading()) {

            case Constants.NORTH:

                yObs -= OwnSpaceCraft.INCREMENT;

                break;

            case Constants.EAST:

                xObs += OwnSpaceCraft.INCREMENT;

                break;

            case Constants.SOUTH:

                yObs += OwnSpaceCraft.INCREMENT;

                break;

            case Constants.WEST:

                xObs -= OwnSpaceCraft.INCREMENT;

                break;

            default:

                System.err.println("Case error in obstacleClear");

        }

        ListIterator iter = obstacles.listIterator();

        while (iter.hasNext()) {

            Obstacle obs = (Obstacle) iter.next();

            if (Math.abs(obs.getXPosition() - xObs) < 10
                    && Math.abs(obs.getYPosition() - yObs) < 10) {

                // System.out.println( "Obstacle in path");
                return false;
            }
        }

        // System.out.println( "Path clear of obstacles");
        return true;

    } // end obstacleClear


    /**
     * Add a single obstacle to those in the sector. Assumes that is will only be
     * called once per obstacle. Does not check for duplicates.
     *
     * @param obs obstacle to be added to the scene
     */
    public synchronized void addObstacle(Obstacle obs) {
        obstacles.add(obs);

        return;

    } // end addObstacle


    /**
     * Remove a SpaceCraft from the sector. The SpaceCraft to be removed could be
     * either alien or ownShip.
     *
     * @param craft SpaceCraft to be removed.
     */
    public synchronized void removeSpaceCraft(SpaceCraft craft) {
        SpaceCraft target;
        ListIterator iter = inSector.listIterator();

        if (ownShip != null && craft.equals(ownShip)) {

            ownShip = null;
        } else {

            while (iter.hasNext()) {

                target = (SpaceCraft) iter.next();

                if (craft.equals(target)) {

                    inSector.remove(iter.previousIndex());

                    return;
                }
            }
        }

    } // end remove


    /**
     * Creates stars in random positions for the display
     */
    public void createStars() {
        for (int i = 0; i < starPositions.length; i++) {

            starPositions[i] = new Point();

            starPositions[i].x = rand.nextInt(sizeX);
            starPositions[i].y = rand.nextInt(sizeY);
        }

    } // end createStars


    /**
     * Renders all the stars in the sector.
     *
     * @param g Graphics context for rendering
     */
    public void drawStars(Graphics g) {
        g.setColor(Color.WHITE);

        for (int i = 0; i < starPositions.length; i++) {

            g.fillOval(starPositions[i].x, starPositions[i].y, 2, 2);
        }
    }


    /**
     * Animation thread. Sleeps inbetween frames according to the frame rate.
     */
    public void run() {

        //Just to be nice, lower this thread's priority
        //so it can't interfere with other processing going on.
        Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

        //Remember the starting time.
        long startTime = System.currentTimeMillis();

        //Remember which thread we are.
        Thread currentThread = Thread.currentThread();

        //This is the animation loop.
        while (currentThread == animatorThread) {

            repaint();

            //Delay depending on how far we are behind.

            try {
                startTime += delay;
                Thread.sleep(Math.max(0, startTime - System.currentTimeMillis()));
            }
            catch (InterruptedException e) {
                break;
            }
        }
    } // end run


    /**
     * Override of the paint method for double buffering.
     */
    public void paint(Graphics g) {
        update(g);

    } // end paint


    /**
     * Renders all object in the sector to the back buffer and then swaps buffers.
     */
    public void update(Graphics g) {
        if (this.getBackground() != null) offGraphics.setColor(getBackground());
        offGraphics.fillRect(0, 0, offDimension.width, offDimension.height);

        // Call the draw methods for all objects in the sector
        drawStars(offGraphics);

        drawObstacles(offGraphics);

        drawAllSpaceCraft(offGraphics);

        g.drawImage(offImage, 0, 0, this);

    } // end update


    /**
     * Draws all alien space craft in the sector. Draws own ship if it is present.
     *
     * @param g Graphics context for rendering
     */
    public synchronized void drawAllSpaceCraft(Graphics g) {
        ListIterator iter = inSector.listIterator();

        while (iter.hasNext()) {

            ((SpaceCraft) (iter.next())).draw(offGraphics);

        }

        if (this.ownShip != null) {

            ownShip.draw(g);

        }

    } // end drawAllSpaceCraft


    /**
     * Draws all obstacles in the sector.
     *
     * @param g Graphics context for rendering
     */
    public synchronized void drawObstacles(Graphics g) {
        ListIterator iter = obstacles.listIterator();

        while (iter.hasNext()) {

            ((Obstacle) (iter.next())).draw(offGraphics);
        }

    } // end drawObstacles


    /**
     * Creates a space craft for the user in a random position with a random
     * heading.
     *
     * @return reference to the users space craft for updating
     */
    public synchronized OwnSpaceCraft createOwnSpaceCraft() {
        if (ownShip == null) {
            ownShip = new OwnSpaceCraft(ownPort);
        }

        return ownShip;
    } // end createOwnSpaceCraft


    /**
     * @author bachmaer
     *         <p/>
     *         Inner class for handling keyboard input.
     */
    class keyboardHandler extends KeyAdapter {
        public void keyPressed(KeyEvent e) {

            switch (e.getKeyCode()) {

                case KeyEvent.VK_UP:

                    System.out.println("Up arrow");

                    break;

                case KeyEvent.VK_DOWN:

                    System.out.println("Down arrow");

                    break;

                default:

                    System.out.println("Pressed key = " + e.getKeyChar());
                    break;
            }
        } // end keyPressed


    } // end keyboardHandler inner class


    /**
     * @author bachmaer
     *         <p/>
     *         Inner class for handling size change events
     */
    class sizeChangeListener extends ComponentAdapter {
        public void componentResized(ComponentEvent e) {

            Dimension newSize = e.getComponent().getSize();

            System.out.println("Width = " + newSize.width
                    + " Height = " + newSize.height);

            newSize.width = (newSize.width > 0) ? newSize.width : 1;
            newSize.height = (newSize.height > 0) ? newSize.height : 1;

            if ((offGraphics == null)
                    || (newSize.width != offDimension.width)
                    || (newSize.height != offDimension.height)) {

                offImage = createImage(newSize.width, newSize.height);

                offGraphics = offImage.getGraphics();
                offGraphics.setColor(getBackground());
                offDimension = newSize;
            }
        }
    } // end sizeChangeListener inner class


    /**
     * main for testing the Sector class.
     *
     * @param args
     */
    public static void main(String[] args) {
        Frame f = new Frame("sector test");

        Sector sector = new Sector(5000);
        f.add(sector);

        f.setSize(600, 600);

        f.setVisible(true);

        sector.updateAndAddSpaceCraft(new OwnSpaceCraft(5000));

    } // end main


} // end Sector class

