import javax.swing.*;
import java.net.DatagramSocket;
import java.net.Socket;
import java.net.InetAddress;
import java.io.DataInputStream;

/**
 * @author bachmaer
 *         <p/>
 *         Driver class for a simple networked space game. Opponents try to destroy each
 *         other by ramming. Head on collisions destroy both ships. Ships move and turn
 *         through GUI mouse clicks. All friendly and alien ships are displayed on a 2D
 *         interface.
 */
public class SpaceGame {
    // Keeps track of the game state
    static Sector sector;

    // User interface
    SpaceGui gui;

    // Port identifiy ownship
    int ownPort;

    private int serverPort = 4444;

    UDPSender sender;


    /**
     * Creates all components needed to start a space game. Creates Sector
     * canvas, GUI interface, a Sender obeject for sending update messages, a
     * UDPReciever object for receiving messages.
     */
    public SpaceGame() {
        try {
            InetAddress server = InetAddress.getByName(JOptionPane.showInputDialog("Please enter a valid server to connect to (hit cancel for localhost):"));
            // InetAddress server = InetAddress.getLocalHost();

            // Create UDP Datagram Socket
            DatagramSocket dgSock = new DatagramSocket();

            sender = new UDPSender(dgSock, server, serverPort);

            // Assign the port number for the Datagram Socket being used to
            // send messages to this.ownPort
            this.ownPort = dgSock.getLocalPort();

            // Create display, ownPort is used to uniquely identify the
            // controlled entity.
            sector = new Sector(ownPort);

            //	gui will call SpaceGame methods to handle user events
            gui = new SpaceGui(this, sector);

            // Use UDP/IP or TCP/IP to receive obstacles from the server
            Socket tcpSock = new Socket(server, serverPort);
            DataInputStream dis = new DataInputStream(tcpSock.getInputStream());
            int i;
            while ((i = dis.readInt()) != -1)
                sector.addObstacle(new Obstacle(i, dis.readInt()));
            dis.close();
            tcpSock.close();

            // Infinite loop or separate thread to receive update messages from the server
            new UDPReciever(dgSock).start();

        } catch (Exception e) {
            e.printStackTrace();
        }


    } // end SpaceGame constructor


    /**
     * Causes sector.ownShip to turn and sends an update message for the heading
     * change.
     */
    public void turnRight() {
        if (sector.ownShip != null) {
            System.out.println(" Right Turn ");

            // Update the display
            sector.ownShip.rightTurn();

            // Send update message to server with new heading.
            sender.update();
        }

    } // end turnRight


    /**
     * Causes sector.ownShip to turn and sends an update message for the heading
     * change.
     */
    public void turnLeft() {
        // See if the player has a ship in play
        if (sector.ownShip != null) {
            System.out.println(" Left Turn ");

            // Update the display
            sector.ownShip.leftTurn();

            // Send update message to other server with new heading.
            sender.update();
        }

    } // end turnLeft


    /**
     * Causes sector.ownShip to move forward and sends an update message for the
     * position change. Checks for a collision with an alien ship in the sector.
     * If collision occurs it sends a remove message for the alien impacted. If
     * the collision was head on the sector.ownShip will become null and a remove
     * message is sent for the sector.ownShip. If there is an obstacle in front of
     * the ship it will not move forward and a message is not sent.
     */
    public void moveFoward() {
        // Check if the player has and unblocked ship in the game
        if (sector.ownShip != null && sector.obstacleClear()) {

            System.out.println(" Move Forward");

            //Update the displayed posistion of the ship
            sector.ownShip.move();

            // Send a message with the updated position to server
            sender.update();

            // Save a reference to ownShip in case it is destroyed in a head
            // on collision.
            OwnSpaceCraft self = sector.ownShip;

            SpaceCraft target = sector.collisionCheck();

            if (target != null) {
                // Send a remove to server for the ship that was rammed
                sender.remove(target);

                // Check if own ship was destroyed in a collision
                if (sector.ownShip == null) {

                    // Send a remove message for own ship to server
                    sender.remove(self);

                }
            }
        }

    } // end moveFoward


    /**
     * Creates a new sector.ownShip if one does not exist. Sends a join message
     * for the new ship.
     */
    public void join() {
        if (sector.ownShip == null) {

            //System.out.println( " Join " );

            // Add a new ownShip to the sector display
            sector.createOwnSpaceCraft();

            // Send message to server let them know you have joined the game using the
            // send object
            sender.join();
        }
    } // end join


    /*
      * Starts the space game. Driver for the application.
      */
    public static void main(String[] args) {
        new SpaceGame();

    } // end main


} // end SpaceGame class
